/*
  Copyright (c) 2024 Tom Goff

  ESP32 Power Data Logger with Display and Bluetooth Classic.
  ================================================================
*/

#include <PZEM004Tv30.h>
#include <RTClib.h> // for the RTC
#include "BluetoothSerial.h"
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <Adafruit_SH110X.h>
#include <SPI.h> //for the SD card module
#include <SD.h> // for the SD card


#if !defined(CONFIG_BT_ENABLED) || !defined(CONFIG_BLUEDROID_ENABLED)
#error Bluetooth is not enabled! Please run `make menuconfig` to and enable it
#endif

#define reset_pin   33

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels

//Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, -1);
Adafruit_SH1106G display = Adafruit_SH1106G(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, -1);

BluetoothSerial SerialBT;

const int chipSelect = 5;

// Create a file to store the data
File myFile;

// RTC
RTC_DS1307 rtc;


PZEM004Tv30 pzem(Serial2, 16, 17);

void IRAM_ATTR resetEnergy()
{
  Serial.println ("reset");
  pzem.resetEnergy();
}

void setup() {
  /* Debugging serial */
  Serial.begin(57600);

#ifndef ESP8266
  while (!Serial); // wait for serial port to connect. Needed for native USB
#endif

  SerialBT.begin("ESP32test"); //Bluetooth device name
  // bt.begin(9600);removed when moving to ESP32

  display.begin(0x3c, true); // Address 0x3C default
  //  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) { // Address 0x3D for 128x64
  //    Serial.println(F("SSD1306 allocation failed"));
  //    for (;;);
  //  }
  delay(2000);
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(WHITE);

  if (! rtc.begin()) {
    Serial.println("Couldn't find RTC");
    Serial.flush();
    while (1) delay(10);
  }

  if (! rtc.isrunning()) {
    Serial.println("RTC is NOT running, let's set the time!");
    // When time needs to be set on a new device, or after a power loss, the
    // following line sets the RTC to the date & time this sketch was compiled
    rtc.adjust(DateTime(F(__DATE__), F(__TIME__)));
    // This line sets the RTC with an explicit date & time, for example to set
    // January 21, 2014 at 3am you would call:
    // rtc.adjust(DateTime(2014, 1, 21, 3, 0, 0));
  }

  rtc.adjust(DateTime(F(__DATE__), F(__TIME__))); // adjust date and time anyway??

  // setup for the SD card
  Serial.print("Initializing SD card...");

  if (!SD.begin(chipSelect)) {
    Serial.println("initialization failed!");
    return;
  }
  Serial.println("initialization done.");

  //open file
  myFile = SD.open("/DATA.txt", FILE_WRITE);

  // if the file opened ok, write to it:
  if (myFile) {
    Serial.println("File opened ok");
    // print the headings for our data
    myFile.println("Date,Time,Power");
  }
  myFile.close();

  pinMode(reset_pin, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(reset_pin), resetEnergy, RISING);
}

void loop() {
  display.clearDisplay();
  Serial.print("Custom Address:");
  Serial.println(pzem.readAddress(), HEX);

  // Read the data from the sensor
  float voltage = pzem.voltage();
  float current = pzem.current();
  float power = pzem.power();
  float energy = pzem.energy();
  float frequency = pzem.frequency();
  float pf = pzem.pf();

  // Check if the data is valid
  if (isnan(voltage)) {
    Serial.println("Error reading voltage");
  } else if (isnan(current)) {
    Serial.println("Error reading current");
  } else if (isnan(power)) {
    Serial.println("Error reading power");
  } else if (isnan(energy)) {
    Serial.println("Error reading energy");
  } else if (isnan(frequency)) {
    Serial.println("Error reading frequency");
  } else if (isnan(pf)) {
    Serial.println("Error reading power factor");
  } else {

    logData(voltage, current, power, energy);
    displayData(voltage, current, power, energy);
    BTData(voltage, current, power, energy);

    // Print the values to the Serial console
    Serial.print("Voltage: ");      Serial.print(voltage);      Serial.println("V");
    Serial.print("Current: ");      Serial.print(current);      Serial.println("A");
    Serial.print("Power: ");        Serial.print(power);        Serial.println("W");
    Serial.print("Energy: ");       Serial.print(energy, 3);     Serial.println("kWh");
    Serial.print("Frequency: ");    Serial.print(frequency, 1); Serial.println("Hz");
    Serial.print("PF: ");           Serial.println(pf);

  }

  Serial.println();
  delay(2000);
}


void logData (float voltage, float current, float power, float energy) {
  DateTime now = rtc.now();
  myFile = SD.open("/DATA.txt", FILE_APPEND);
  if (myFile) {
    myFile.print(now.year(), DEC);
    myFile.print('/');
    myFile.print(now.month(), DEC);
    myFile.print('/');
    myFile.print(now.day(), DEC);
    myFile.print(',');
    myFile.print(now.hour(), DEC);
    myFile.print(':');
    myFile.print(now.minute(), DEC);
    myFile.print(':');
    myFile.print(now.second(), DEC);
    myFile.print(" - ");
    myFile.print(energy);
    myFile.print("kWh");
    myFile.println(",");
    myFile.close();
    Serial.println("data saves");
  }
  Serial.print(now.year(), DEC);
  Serial.print('/');
  Serial.print(now.month(), DEC);
  Serial.print('/');
  Serial.println(now.day(), DEC);
  Serial.print(now.hour(), DEC);
  Serial.print(':');
  Serial.print(now.minute(), DEC);
  Serial.print(':');
  Serial.println(now.second(), DEC);

  delay(1000);
}

void displayData (float voltage, float current, float power, float energy) {
  float power_kw = power / 1000;
  display.setCursor(0, 10);
  display.println("Voltage: ");
  display.display();
  display.setCursor(110, 10);
  display.println("V");
  display.display();
  display.setCursor(65, 10);
  display.println(voltage);
  display.display();

  display.setCursor(0, 20);
  display.println("Current: ");
  display.display();
  display.setCursor(110, 20);
  display.println("A");
  display.display();
  display.setCursor(65, 20);
  display.println(current);
  display.display();

  display.setCursor(0, 30);
  display.println("Power: ");
  display.display();
  display.setCursor(110, 30);
  display.println("kW");
  display.display();
  display.setCursor(65, 30);
  display.println(power_kw);
  display.display();

  display.setCursor(0, 40);
  display.println("Energy: ");
  display.display();
  display.setCursor(110, 40);
  display.println("kWh");
  display.display();
  display.setCursor(65, 40);
  display.println(energy);
  display.display();
}

void BTData (float voltage, float current, float power, float energy) {
  float power_kw = power / 1000;
  SerialBT.print(voltage); //send distance to MIT App
  SerialBT.print("|"); // to separate values in the phone app
  SerialBT.print(current); //send distance to MIT App
  SerialBT.print("|");
  SerialBT.print(power_kw); //send distance to MIT App
  SerialBT.print("|");
  SerialBT.print(energy); //send distance to MIT App

}
